/*
 * __get_user functions.
 *
 * (C) Copyright 1998 Linus Torvalds
 *
 * These functions have a non-standard call interface
 * to make them more efficient, especially as they
 * return an error value in addition to the "real"
 * return value.
 */
#include <asm/thread_info.h>


/*
 * __get_user_X
 *
 * Inputs:	%eax contains the address
 *
 * Outputs:	%eax is error code (0 or -EFAULT)
 *		%edx contains zero-extended value
 *
 * These functions should not modify any other registers,
 * as they get called from within inline assembly.
 */

.text
.align 4
.globl __get_user_1
__get_user_1:
	GET_THREAD_INFO(%edx)
	cmpl TI_addr_limit(%edx),%eax
	jae bad_get_user
#ifdef CONFIG_KRG_FAF
	testl $_TIF_RUACCESS,TI_flags(%edx)
	jnz 10f
#endif
1:	movzbl (%eax),%edx
	xorl %eax,%eax
	ret
#ifdef CONFIG_KRG_FAF
10:
	cmpl $-1,TI_addr_limit(%edx)
	je 1b
	movl $1,%edx
	jmp remote_get_user
#endif

.align 4
.globl __get_user_2
__get_user_2:
	addl $1,%eax
	jc bad_get_user
	GET_THREAD_INFO(%edx)
	cmpl TI_addr_limit(%edx),%eax
	jae bad_get_user
#ifdef CONFIG_KRG_FAF
	testl $_TIF_RUACCESS,TI_flags(%edx)
	jnz 20f
#endif
2:	movzwl -1(%eax),%edx
	xorl %eax,%eax
	ret
#ifdef CONFIG_KRG_FAF
20:
	cmpl $-1,TI_addr_limit(%edx)
	je 2b
	decl %eax
	movl $2,%edx
	jmp remote_get_user
#endif

.align 4
.globl __get_user_4
__get_user_4:
	addl $3,%eax
	jc bad_get_user
	GET_THREAD_INFO(%edx)
	cmpl TI_addr_limit(%edx),%eax
	jae bad_get_user
#ifdef CONFIG_KRG_FAF
	testl $_TIF_RUACCESS,TI_flags(%edx)
	jnz 30f
#endif
3:	movl -3(%eax),%edx
	xorl %eax,%eax
	ret
#ifdef CONFIG_KRG_FAF
30:
	cmpl $-1,TI_addr_limit(%edx)
	je 3b
	subl $3,%eax
	movl $4,%edx
	jmp remote_get_user
#endif

bad_get_user:
	xorl %edx,%edx
	movl $-14,%eax
	ret

.section __ex_table,"a"
	.long 1b,bad_get_user
	.long 2b,bad_get_user
	.long 3b,bad_get_user
.previous

#ifdef CONFIG_KRG_FAF
remote_get_user:
	/*
	 * gcc does not know that we clobber ebx and ecx,
	 * and we don't really care about the cost of saving registers here,
	 * since we are going to wait for the result from a remote host.
	 */
	pushl %ebx
	pushl %ecx
	subl $8,%esp
	movl kh_copy_user_generic,%ebx
	movl $0,(%esp)
	movl %edx,%ecx
	movl %eax,%edx
	leal 4(%esp),%eax
	call *%ebx
	addl $4,%esp
	popl %edx
	popl %ecx
	popl %ebx
	testl %eax,%eax
	jnz bad_get_user
	ret
#endif /* CONFIG_KRG_FAF */
