/*
 * __put_user functions.
 *
 * (C) Copyright 2005 Linus Torvalds
 *
 * These functions have a non-standard call interface
 * to make them more efficient, especially as they
 * return an error value in addition to the "real"
 * return value.
 */
#include <asm/thread_info.h>


/*
 * __put_user_X
 *
 * Inputs:	%eax[:%edx] contains the data
 *		%ecx contains the address
 *
 * Outputs:	%eax is error code (0 or -EFAULT)
 *
 * These functions should not modify any other registers,
 * as they get called from within inline assembly.
 */

#define ENTER	pushl %ebx ; GET_THREAD_INFO(%ebx)
#define EXIT	popl %ebx ; ret

.text
.align 4
.globl __put_user_1
__put_user_1:
	ENTER
	cmpl TI_addr_limit(%ebx),%ecx
	jae bad_put_user
#ifdef CONFIG_KRG_FAF
	testl $_TIF_RUACCESS,TI_flags(%ebx)
	jnz 10f
#endif
1:	movb %al,(%ecx)
	xorl %eax,%eax
	EXIT
#ifdef CONFIG_KRG_FAF
10:
	cmpl $-1,TI_addr_limit(%ebx)
	je 1b
	movl $1,%ebx
	jmp remote_put_user
#endif

.align 4
.globl __put_user_2
__put_user_2:
	ENTER
#ifdef CONFIG_KRG_FAF
	testl $_TIF_RUACCESS,TI_flags(%ebx)
	jnz 20f
200:
#endif
	movl TI_addr_limit(%ebx),%ebx
	subl $1,%ebx
	cmpl %ebx,%ecx
	jae bad_put_user
2:	movw %ax,(%ecx)
	xorl %eax,%eax
	EXIT
#ifdef CONFIG_KRG_FAF
20:
	cmpl $-1,TI_addr_limit(%ebx)
	je 200b
	movl TI_addr_limit(%ebx),%ebx
	subl $1,%ebx
	cmpl %ebx,%ecx
	jae bad_put_user
	movl $2,%ebx
	jmp remote_put_user
#endif

.align 4
.globl __put_user_4
__put_user_4:
	ENTER
#ifdef CONFIG_KRG_FAF
	testl $_TIF_RUACCESS,TI_flags(%ebx)
	jnz 30f
300:
#endif
	movl TI_addr_limit(%ebx),%ebx
	subl $3,%ebx
	cmpl %ebx,%ecx
	jae bad_put_user
3:	movl %eax,(%ecx)
	xorl %eax,%eax
	EXIT
#ifdef CONFIG_KRG_FAF
30:
	cmpl $-1,TI_addr_limit(%ebx)
	je 300b
	movl TI_addr_limit(%ebx),%ebx
	subl $3,%ebx
	cmpl %ebx,%ecx
	jae bad_put_user
	movl $4,%ebx
	jmp remote_put_user
#endif

.align 4
.globl __put_user_8
__put_user_8:
	ENTER
#ifdef CONFIG_KRG_FAF
	testl $_TIF_RUACCESS,TI_flags(%ebx)
	jnz 40f
400:
#endif
	movl TI_addr_limit(%ebx),%ebx
	subl $7,%ebx
	cmpl %ebx,%ecx
	jae bad_put_user
4:	movl %eax,(%ecx)
5:	movl %edx,4(%ecx)
	xorl %eax,%eax
	EXIT
#ifdef CONFIG_KRG_FAF
40:
	cmpl $-1,TI_addr_limit(%ebx)
	je 400b
	movl TI_addr_limit(%ebx),%ebx
	subl $7,%ebx
	cmpl %ebx,%ecx
	jae bad_put_user
	movl $8,%ebx
	jmp remote_put_user
#endif

bad_put_user:
	movl $-14,%eax
	EXIT

.section __ex_table,"a"
	.long 1b,bad_put_user
	.long 2b,bad_put_user
	.long 3b,bad_put_user
	.long 4b,bad_put_user
	.long 5b,bad_put_user
.previous

#ifdef CONFIG_KRG_FAF
remote_put_user:
	/*
	 * gcc does not know that we clobber ecx and edx,
	 * and we don't really care about the cost of saving registers here,
	 * since we are going to wait for the result from a remote host.
	 */
	pushl %ecx
	pushl %edx
	pushl %eax
	movl %ecx,%eax
	movl %ebx,%ecx
	movl kh_copy_user_generic,%ebx
	movl %esp,%edx
	subl $4,%esp
	movl $0,(%esp)
	call *%ebx
	addl $8,%esp
	popl %edx
	popl %ecx
	testl %eax,%eax
	jnz bad_put_user
	EXIT
#endif /* CONFIG_KRG_FAF */
