/*
 * __get_user functions.
 *
 * (C) Copyright 1998 Linus Torvalds
 * (C) Copyright 2005 Andi Kleen
 *
 * These functions have a non-standard call interface
 * to make them more efficient, especially as they
 * return an error value in addition to the "real"
 * return value.
 */

/*
 * __get_user_X
 *
 * Inputs:	%rcx contains the address.
 *		The register is modified, but all changes are undone
 *		before returning because the C code doesn't know about it.
 *
 * Outputs:	%rax is error code (0 or -EFAULT)
 *		%rdx contains zero-extended value
 * 
 * %r8 is destroyed.
 *
 * These functions should not modify any other registers,
 * as they get called from within inline assembly.
 */

#include <linux/linkage.h>
#include <asm/dwarf2.h>
#include <asm/page.h>
#include <asm/errno.h>
#include <asm/asm-offsets.h>
#include <asm/thread_info.h>

	.text
ENTRY(__get_user_1)
	CFI_STARTPROC
	GET_THREAD_INFO(%r8)
	cmpq threadinfo_addr_limit(%r8),%rcx
	jae bad_get_user
1:	movzb (%rcx),%edx
	xorl %eax,%eax
	ret
#ifdef CONFIG_KRG_FAF
100:
	movq $1,%rdx
	jmp ruaccess_get_user
#endif
	CFI_ENDPROC
ENDPROC(__get_user_1)

ENTRY(__get_user_2)
	CFI_STARTPROC
	GET_THREAD_INFO(%r8)
	addq $1,%rcx
	jc 20f
	cmpq threadinfo_addr_limit(%r8),%rcx
	jae 20f
	decq   %rcx
2:	movzwl (%rcx),%edx
	xorl %eax,%eax
	ret
20:	decq    %rcx
	jmp	bad_get_user
#ifdef CONFIG_KRG_FAF
200:
	movq $2,%rdx
	jmp ruaccess_get_user
#endif
	CFI_ENDPROC
ENDPROC(__get_user_2)

ENTRY(__get_user_4)
	CFI_STARTPROC
	GET_THREAD_INFO(%r8)
	addq $3,%rcx
	jc 30f
	cmpq threadinfo_addr_limit(%r8),%rcx
	jae 30f
	subq $3,%rcx
3:	movl (%rcx),%edx
	xorl %eax,%eax
	ret
30:	subq $3,%rcx
	jmp bad_get_user
#ifdef CONFIG_KRG_FAF
300:
	movq $4,%rdx
	jmp ruaccess_get_user
#endif
	CFI_ENDPROC
ENDPROC(__get_user_4)

ENTRY(__get_user_8)
	CFI_STARTPROC
	GET_THREAD_INFO(%r8)
	addq $7,%rcx
	jc 40f
	cmpq threadinfo_addr_limit(%r8),%rcx
	jae	40f
	subq	$7,%rcx
4:	movq (%rcx),%rdx
	xorl %eax,%eax
	ret
40:	subq $7,%rcx
	jmp bad_get_user
#ifdef CONFIG_KRG_FAF
400:
	movq $8,%rdx
	jmp ruaccess_get_user
#endif
	CFI_ENDPROC
ENDPROC(__get_user_8)

bad_get_user:
	CFI_STARTPROC
	xorl %edx,%edx
	movq $(-EFAULT),%rax
	ret
	CFI_ENDPROC
END(bad_get_user)

.section __ex_table,"a"
#ifdef CONFIG_KRG_FAF
	.quad 1b,100b
	.quad 2b,200b
	.quad 3b,300b
	.quad 4b,400b
#else /* CONFIG_KRG_FAF */
	.quad 1b,bad_get_user
	.quad 2b,bad_get_user
	.quad 3b,bad_get_user
	.quad 4b,bad_get_user
#endif /* CONFIG_KRG_FAF */
.previous

#ifdef CONFIG_KRG_FAF
ruaccess_get_user:
	CFI_STARTPROC
	testl $_TIF_RUACCESS,threadinfo_flags(%r8)
	jz bad_get_user
	cmpq $-1,threadinfo_addr_limit(%r8)
	je bad_get_user
	/*
	 * gcc does not know that we (may) clobber rdi,rsi,rcx,r9,r10,r11,
	 * and we don't really care about the cost of saving registers here,
	 * since we are going to wait for the result from a remote host.
	 */
	pushq %rdi
	CFI_ADJUST_CFA_OFFSET	8
	pushq %rsi
	CFI_ADJUST_CFA_OFFSET	8
	pushq %rcx
	CFI_ADJUST_CFA_OFFSET	8
	pushq %r9
	CFI_ADJUST_CFA_OFFSET	8
	pushq %r10
	CFI_ADJUST_CFA_OFFSET	8
	pushq %r11
	CFI_ADJUST_CFA_OFFSET	8
	/* %rdx is zero-filled on success */
	xorl %eax,%eax
	pushq %rax
	CFI_ADJUST_CFA_OFFSET	8
	movq kh_copy_user_generic,%rax
	movq %rsp,%rdi
	movq %rcx,%rsi
	/* %rdx is set by the caller */
	xorl %ecx,%ecx
	callq *%rax
	/* shorter form of 'movq (%rsp),%rdx' */
	popq %rdx
	CFI_ADJUST_CFA_OFFSET	-8
	popq %r11
	CFI_ADJUST_CFA_OFFSET	-8
	popq %r10
	CFI_ADJUST_CFA_OFFSET	-8
	popq %r9
	CFI_ADJUST_CFA_OFFSET	-8
	popq %rcx
	CFI_ADJUST_CFA_OFFSET	-8
	popq %rsi
	CFI_ADJUST_CFA_OFFSET	-8
	popq %rdi
	CFI_ADJUST_CFA_OFFSET	-8
	testl %eax,%eax
	jnz bad_get_user
	ret
	CFI_ENDPROC
END(ruaccess_get_user)

/*
 * Inputs:	%rax contains the address to copy from
 *		24(%rsp) is the number of bytes to read
 *
 * Outpus:	%rax contains 0 if a value was successfully read remotely, or
 *		the number of bytes that were failed to read (either because no
 *		remote access was setup, or because read failed remotely).
 *		16(%rsp) contains the remotely read value on success.
 *
 * All (possibly) clobbered registers but %rax are saved since we are called
 * from inline assembly, and don't want to slow down the fast path with added
 * clobbers.
 */
ENTRY(ruaccess_get_user_asm)
	CFI_STARTPROC
	pushq %rdi
	CFI_ADJUST_CFA_OFFSET	8
	pushq %rdx
	CFI_ADJUST_CFA_OFFSET	8
	GET_THREAD_INFO(%rdi)
	/* 24 + 2*8 = 40 */
	movq 40(%rsp),%rdx
	testl $_TIF_RUACCESS,threadinfo_flags(%rdi)
	jz 6f
	cmpq $-1,threadinfo_addr_limit(%rdi)
	je 6f
	pushq %rsi
	CFI_ADJUST_CFA_OFFSET	8
	pushq %rcx
	/* kh_copy_user_generic may clobber r8-r11 */
	CFI_ADJUST_CFA_OFFSET	8
	pushq %r8
	CFI_ADJUST_CFA_OFFSET	8
	pushq %r9
	CFI_ADJUST_CFA_OFFSET	8
	pushq %r10
	CFI_ADJUST_CFA_OFFSET	8
	pushq %r11
	CFI_ADJUST_CFA_OFFSET	8
	/* 16 + 2*8 + 6*8 = 80 */
	leaq 80(%rsp),%rdi
	movq %rax,%rsi
	/* %rdx is already loaded */
	xorl %ecx,%ecx
	movq kh_copy_user_generic,%rax
	call *%rax
	popq %r11
	CFI_ADJUST_CFA_OFFSET	-8
	popq %r10
	CFI_ADJUST_CFA_OFFSET	-8
	popq %r9
	CFI_ADJUST_CFA_OFFSET	-8
	popq %r8
	CFI_ADJUST_CFA_OFFSET	-8
	popq %rcx
	CFI_ADJUST_CFA_OFFSET	-8
	popq %rsi
	CFI_ADJUST_CFA_OFFSET	-8
5:
	popq %rdx
	CFI_ADJUST_CFA_OFFSET	-8
	popq %rdi
	CFI_ADJUST_CFA_OFFSET	-8
	ret
6:
	movq %rdx,%rax
	jmp 5b
	CFI_ENDPROC
END(ruaccess_get_user_asm)

#endif /* CONFIG_KRG_FAF */
