/*
 * __put_user functions.
 *
 * (C) Copyright 1998 Linus Torvalds
 * (C) Copyright 2005 Andi Kleen
 *
 * These functions have a non-standard call interface
 * to make them more efficient, especially as they
 * return an error value in addition to the "real"
 * return value.
 */

/*
 * __put_user_X
 *
 * Inputs:	%rcx contains the address
 *		%rdx contains new value
 *
 * Outputs:	%rax is error code (0 or -EFAULT)
 *
 * %r8 is destroyed.
 *
 * These functions should not modify any other registers,
 * as they get called from within inline assembly.
 */

#include <linux/linkage.h>
#include <asm/dwarf2.h>
#include <asm/page.h>
#include <asm/errno.h>
#include <asm/asm-offsets.h>
#include <asm/thread_info.h>

	.text
ENTRY(__put_user_1)
	CFI_STARTPROC
	GET_THREAD_INFO(%r8)
	cmpq threadinfo_addr_limit(%r8),%rcx
	jae bad_put_user
1:	movb %dl,(%rcx)
	xorl %eax,%eax
	ret
#ifdef CONFIG_KRG_FAF
100:	movq $1,%rax
	jmp ruaccess_put_user
#endif
	CFI_ENDPROC
ENDPROC(__put_user_1)

ENTRY(__put_user_2)
	CFI_STARTPROC
	GET_THREAD_INFO(%r8)
	addq $1,%rcx
	jc 20f
	cmpq threadinfo_addr_limit(%r8),%rcx
	jae 20f
	decq %rcx
2:	movw %dx,(%rcx)
	xorl %eax,%eax
	ret
20:	decq %rcx
	jmp bad_put_user
#ifdef CONFIG_KRG_FAF
200:	movq $2,%rax
	jmp ruaccess_put_user
#endif
	CFI_ENDPROC
ENDPROC(__put_user_2)

ENTRY(__put_user_4)
	CFI_STARTPROC
	GET_THREAD_INFO(%r8)
	addq $3,%rcx
	jc 30f
	cmpq threadinfo_addr_limit(%r8),%rcx
	jae 30f
	subq $3,%rcx
3:	movl %edx,(%rcx)
	xorl %eax,%eax
	ret
30:	subq $3,%rcx
	jmp bad_put_user
#ifdef CONFIG_KRG_FAF
300:	movq $4,%rax
	jmp ruaccess_put_user
#endif
	CFI_ENDPROC
ENDPROC(__put_user_4)

ENTRY(__put_user_8)
	CFI_STARTPROC
	GET_THREAD_INFO(%r8)
	addq $7,%rcx
	jc 40f
	cmpq threadinfo_addr_limit(%r8),%rcx
	jae 40f
	subq $7,%rcx
4:	movq %rdx,(%rcx)
	xorl %eax,%eax
	ret
40:	subq $7,%rcx
	jmp bad_put_user
#ifdef CONFIG_KRG_FAF
400:	movq $8,%rax
	jmp ruaccess_put_user
#endif
	CFI_ENDPROC
ENDPROC(__put_user_8)

bad_put_user:
	CFI_STARTPROC
	movq $(-EFAULT),%rax
	ret
	CFI_ENDPROC
END(bad_put_user)

.section __ex_table,"a"
#ifdef CONFIG_KRG_FAF
	.quad 1b,100b
	.quad 2b,200b
	.quad 3b,300b
	.quad 4b,400b
#else /* CONFIG_KRG_FAF */
	.quad 1b,bad_put_user
	.quad 2b,bad_put_user
	.quad 3b,bad_put_user
	.quad 4b,bad_put_user
#endif /* CONFIG_KRG_FAF */
.previous

#ifdef CONFIG_KRG_FAF
ruaccess_put_user:
	CFI_STARTPROC
	testl $_TIF_RUACCESS,threadinfo_flags(%r8)
	jz bad_put_user
	cmpq $-1,threadinfo_addr_limit(%r8)
	je bad_put_user
	/*
	 * gcc does not know that we (may) clobber rdi,rsi,rdx,rcx,r9,r10,r11,
	 * and we don't really care about the cost of saving registers here,
	 * since we are going to wait for the result from a remote host.
	 */
	pushq %rdi
	CFI_ADJUST_CFA_OFFSET	8
	pushq %rsi
	CFI_ADJUST_CFA_OFFSET	8
	pushq %rcx
	CFI_ADJUST_CFA_OFFSET	8
	pushq %r9
	CFI_ADJUST_CFA_OFFSET	8
	pushq %r10
	CFI_ADJUST_CFA_OFFSET	8
	pushq %r11
	CFI_ADJUST_CFA_OFFSET	8
	pushq %rdx
	CFI_ADJUST_CFA_OFFSET	8
	/* The value to write is at %rsp */
	movq kh_copy_user_generic,%r8
	movq %rcx,%rdi
	movq %rsp,%rsi
	movq %rax,%rdx
	xorl %ecx,%ecx
	callq *%r8
	popq %rdx
	CFI_ADJUST_CFA_OFFSET	-8
	popq %r11
	CFI_ADJUST_CFA_OFFSET	-8
	popq %r10
	CFI_ADJUST_CFA_OFFSET	-8
	popq %r9
	CFI_ADJUST_CFA_OFFSET	-8
	popq %rcx
	CFI_ADJUST_CFA_OFFSET	-8
	popq %rsi
	CFI_ADJUST_CFA_OFFSET	-8
	popq %rdi
	CFI_ADJUST_CFA_OFFSET	-8
	testl %eax,%eax
	jnz bad_put_user
	ret
	CFI_ENDPROC
END(ruaccess_put_user)

/*
 * Inputs:	%rax contains the address to copy to
 *		16(%rsp) is the number of bytes to write
 *		8(%rsp) is the value to write
 *
 * Outpus:	%rax contains 0 if a value was successfully written remotely, or
 *		the number of bytes that were failed to write (either because no
 *		remote access was setup, or because write failed remotely).
 *
 * All (possibly) clobbered registers but %rax are saved since we are called
 * from inline assembly, and don't want to slow down the fast path with added
 * clobbers.
 */
ENTRY(ruaccess_put_user_asm)
	CFI_STARTPROC
	pushq %rdi
	CFI_ADJUST_CFA_OFFSET	8
	pushq %rdx
	CFI_ADJUST_CFA_OFFSET	8
	GET_THREAD_INFO(%rdi)
	/* 16 + 2*8 = 32 */
	movq 32(%rsp),%rdx
	testl $_TIF_RUACCESS,threadinfo_flags(%rdi)
	jz 6f
	cmpq $-1,threadinfo_addr_limit(%rdi)
	je 6f
	pushq %rsi
	CFI_ADJUST_CFA_OFFSET	8
	pushq %rcx
	/* kh_copy_user_generic may clobber r8-r11 */
	CFI_ADJUST_CFA_OFFSET	8
	pushq %r8
	CFI_ADJUST_CFA_OFFSET	8
	pushq %r9
	CFI_ADJUST_CFA_OFFSET	8
	pushq %r10
	CFI_ADJUST_CFA_OFFSET	8
	pushq %r11
	CFI_ADJUST_CFA_OFFSET	8
	movq %rax,%rdi
	/* 8 + 2*8 + 6*8 = 72 */
	leaq 72(%rsp),%rsi
	/* %rdx is already loaded */
	xorl %ecx,%ecx
	movq kh_copy_user_generic,%rax
	call *%rax
	popq %r11
	CFI_ADJUST_CFA_OFFSET	-8
	popq %r10
	CFI_ADJUST_CFA_OFFSET	-8
	popq %r9
	CFI_ADJUST_CFA_OFFSET	-8
	popq %r8
	CFI_ADJUST_CFA_OFFSET	-8
	popq %rcx
	CFI_ADJUST_CFA_OFFSET	-8
	popq %rsi
	CFI_ADJUST_CFA_OFFSET	-8
5:
	popq %rdx
	CFI_ADJUST_CFA_OFFSET	-8
	popq %rdi
	CFI_ADJUST_CFA_OFFSET	-8
	ret
6:
	movq %rdx,%rax
	jmp 5b
	CFI_ENDPROC
END(ruaccess_put_user_asm)

#endif /* CONFIG_KRG_FAF */
